/*
 * Pathnames for DragonFly installer backend.
 * $Id: pathnames.h,v 1.2 2004/07/11 06:16:04 cpressey Exp $
 */

#ifndef __PATHNAMES_H_
#define __PATHNAMES_H_

/*
 * Where the files to copy onto the disk reside.
 * Normally this will be "/", when booting from the live CD,
 * but for testing purposes we can set it to (say) "/cdrom/",
 * so that we needn't boot from a CD in development.
 * Note that this must include the trailing slash.
 */
#ifndef DEFAULT_DFLY_ROOT
#define	DEFAULT_DFLY_ROOT	"/"
#endif

/*
 * Temporary directory in which to place disklabels, etc.
 * /tmp/ is MFS mounted when booting from the live CD,
 * so we can use it both in development and in production.
 * Note that this must include the trailing slash.
 */
#ifndef INSTALLER_TEMP
#define	INSTALLER_TEMP	"/tmp/"
#endif

/*
 * File containing the dmesg from boot time.  We look in it
 * to find what disks are attached to this machine.
 * /var/ is MFS mounted when booting from the live CD,
 * so we can use this both in development and in production.
 */
#ifndef DMESG_BOOT
#define	DMESG_BOOT	"/var/run/dmesg.boot"
#endif

/*
 * cpdup command and its arguments.
 */

#ifndef CPDUP_CMD
#define CPDUP_CMD "cpdup -o -vvv -u"
#endif

#endif /* !__PATHNAMES_H_ */
