/*
 * Copyright (c)2004 The DragonFly Project.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 *   Neither the name of the DragonFly Project nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

/*
 * commands.h
 * $Id: commands.h,v 1.1 2004/07/09 23:25:40 cpressey Exp $
 */

#include <stdio.h>

#include <dfui/dfui.h>

#ifndef __COMMANDS_H_
#define __COMMANDS_H_

/*** TYPES ***/

struct commands {
	struct command *head;
	struct command *tail;
};

struct command {
	struct command *next;
	struct command *prev;
	int (*callback)(FILE *, void *);
	void *userdata;
	char *cmd;
	char *desc;
	int logged;
};

/*** PROTOTYPES ***/

struct commands	*commands_new(void);
struct command	*command_new(struct commands *);
struct command	*command_add(struct commands *, char *, ...);
struct command	*command_add_logless(struct commands *, char *, ...);
struct command	*command_add_function(struct commands *,
				      int (*)(FILE *, void *),
				      void *, char *);
int		 command_set_desc(struct command *, char *);
void		 commands_preview(struct commands *);
int		 commands_execute(struct dfui_connection *, struct commands *);
void		 commands_free(struct commands *);

#endif /* !__COMMANDS_H_ */
